import numpy as np

import pycuda.gpuarray as gpuarray
from pycuda.compiler import SourceModule

from pydtnn.layers.layer_normalization import LayerNormalization
from pydtnn.backends.gpu.layers.layer import LayerGPU
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU


class LayerNormalizationGPU(LayerNormalization[TensorGPU], LayerGPU):
    def initialize(self, prev_shape, x):
        super().initialize(prev_shape, x)
        self.shape = prev_shape
        self.x = x
        self.epsilon = np.float32(self.epsilon)

        # Shape same as x input
        self.y = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.y = TensorGPU(self.y, self.model.tensor_fmt, self.model.cudnn_dtype)
        self.dx = gpuarray.empty(x.ary.shape, self.model.dtype)
        self.dx = TensorGPU(self.dx, self.model.tensor_fmt, self.model.cudnn_dtype)

        # Shape same as x input, but batch = 1. For scaling at the end: output = scale * post_normalization + bias
        gamma_shape = (int(np.prod([x.ary.shape[i] for i in self.axis])),)
        self.gamma = gpuarray.to_gpu(np.full(gamma_shape, self.gamma_init_val, self.model.dtype))
        self.gamma = TensorGPU(self.gamma, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)
        self.beta = gpuarray.zeros(gamma_shape, self.model.dtype)
        self.beta = TensorGPU(self.beta, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)
        self.dgamma = gpuarray.zeros(gamma_shape, self.model.dtype)
        self.dgamma = TensorGPU(self.dgamma, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)
        self.dbeta = gpuarray.zeros(gamma_shape, self.model.dtype)
        self.dbeta = TensorGPU(self.dbeta, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)

        # Shape same as x input, but last layer = 1. For mean computation across the normalization axis.
        mean_shape = (int(np.prod(x.ary.shape) / np.prod([x.ary.shape[i] for i in self.axis])),)  # (*x.ary.shape[:-2], 1, 1)
        self.std = gpuarray.empty(mean_shape, self.model.dtype)
        self.std = TensorGPU(self.std, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)
        out_shape = x.ary.shape
        self.xn = gpuarray.empty(out_shape, self.model.dtype)
        self.xn = TensorGPU(self.xn, self.model.tensor_fmt, self.model.cudnn_dtype, tensor_type=TensorGPU.TensorTypeEnum.OTHER)

        self.__init_kernels_gpu__()
        self.threads = int(min(self.kernel_dim_params[0], 1024))
        self.blocks = int(max(self.kernel_dim_params[0], 1024) // self.threads + 1)

        self.threads_backward_weights = int(min(self.kernel_dim_params[1], 1024))
        self.blocks_backward_weights = int(max(self.kernel_dim_params[1], 1024) // self.threads_backward_weights + 1)

    def forward(self, x):
        self.kernel_forward(x.ary, self.y.ary,
                            self.xn.ary, self.std.ary,
                            self.gamma.ary, self.beta.ary,
                            self.epsilon,
                            *self.kernel_dim_params,
                            grid=(self.blocks, 1, 1), block=(self.threads, 1, 1),
                            stream=self.model.stream)
        return self.y

    def backward(self, dy):
        self.kernel_backward(dy.ary, self.dx.ary,
                             self.xn.ary, self.std.ary,
                             self.gamma.ary, self.epsilon,
                             *self.kernel_dim_params,
                             grid=(self.blocks, 1, 1), block=(self.threads, 1, 1),
                             stream=self.model.stream)

        self.kernel_backward_weigths(dy.ary, self.xn.ary,
                                     self.dgamma.ary, self.dbeta.ary, self.epsilon,
                                     *self.kernel_dim_params,
                                     grid=(self.blocks_backward_weights, 1, 1), block=(self.threads_backward_weights, 1, 1),
                                     stream=self.model.stream)

        # print(np.sum(self.beta.ary.get()), np.sum(self.dbeta.ary.get()))
        return self.dx

    def __init_kernels_gpu__(self):
        module = SourceModule("""
        __global__ void gpu_forward(T *x, T *y, T *xn, T *std, T *gamma, T *beta, float epsilon, int batch, int n)
        {
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < batch) {
                int i = 0;
                T mu = 0;
                T var = 0;
                T xc = 0;
                // Mean
                for ( i = 0; i < n; i++ ) {
                    mu += x[idx * n + i] / n;
                }

                // Var
                for ( i = 0; i < n; i++ ){
                    xc = x[idx * n + i] - mu;
                    var += (xc * xc) / n;
                    xn[idx * n + i] = xc;
                }
                var = sqrtf(var + epsilon);
                std[idx] = var;
                // Normalization and Scaling
                for ( i = 0; i < n; i++ ){
                    xn[idx * n + i] /= (var + epsilon);
                    y[idx * n + i] = gamma[i] * xn[idx * n +i] + beta[i];
                }
            }
            return;
        }
        """.replace("T", {np.float32: "float", np.float64: "double"}[self.model.dtype]))
        self.kernel_forward = module.get_function("gpu_forward")
        n = np.prod([self.y.ary.shape[i] for i in self.axis])
        self.kernel_dim_params = (np.int32(np.prod(self.y.ary.shape) // n), np.int32(n))

        module = SourceModule("""
        __global__ void gpu_backward(T *dy, T *dx, T *xn, T *std, T *gamma, float epsilon, int batch, int n)
        {
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < batch) {
                int i = 0;
                T mean1 = 0;
                T mean2 = 0;

                // Means
                for ( i = 0; i < n; i++ ) {
                    mean1 += gamma[i] * xn[idx * n + i] * (dy[idx * n + i] / n);
                    mean2 += gamma[i] * (dy[idx * n + i] / n);
                }

                // dx
                for ( i = 0; i < n; i++ ) {
                    dx[idx * n + i] = (dy[idx * n + i] - xn[idx * n + i] * mean1 - mean2) / (std[idx] + epsilon);
                }
            }
            return;
        }
        """.replace("T", {np.float32: "float", np.float64: "double"}[self.model.dtype]))
        self.kernel_backward = module.get_function("gpu_backward")

        module = SourceModule("""
        __global__ void gpu_backward_weights(T *dy, T *xn, T *dgamma, T *dbeta, float epsilon, int batch, int n)
        {
            int idx = blockIdx.x * blockDim.x + threadIdx.x;
            if (idx < n) {
                int i = 0;
                T mean1 = 0;
                T mean2 = 0;

                // Means
                for ( i = 0; i < batch; i++ ) {
                    mean1 += xn[i * n + idx] * (dy[i * n + idx] / batch);
                    mean2 += dy[i * n + idx] / batch;
                }
                dgamma[idx] = (fabs(mean1) < epsilon) ? 0.0 : mean1;
                dbeta[idx] = (fabs(mean2) < epsilon) ? 0.0 : mean2;
            }
            return;
        }
        """.replace("T", {np.float32: "float", np.float64: "double"}[self.model.dtype]))
        self.kernel_backward_weigths = module.get_function("gpu_backward_weights")

        return



# Typing related (or non important) imports
from typing import Any
from pydtnn.layer_base import LayerBase

# Functionality imports
import pydtnn.layers as layer
import pydtnn.activations as activation
import pydtnn.converters.onnx2pydtnn.constants as cons

# ========================= #

# ===== #
# = A = #
# ===== #


def Add(info: dict[str, Any]) -> LayerBase:

    # TODO: from print to "log - debug" or somthing like that.
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    list_adding_nodes = info[cons.CONST_listS_NODES]

    from pydtnn.layers.addition_block import AdditionBlock
    return AdditionBlock(list_adding_nodes)
# --- END Add --- #


def AveragePool(info: dict[str, Any]) -> LayerBase:

    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__AveragePool.html
    ONNX_COUNT_DILATATIONS = "dilations"
    ONNX_KERNEL_SHAPE = "kernel_shape"
    ONNX_PADS = "pads"
    ONNX_STRIDES = "strides"
    # PyDTNN attributes names from AbstractPool2DLayer class.
    PYDTNN_DILATION = "dilation"
    PYDTNN_POOL_SHAPE = "pool_shape"
    PYDTNN_PADDING = "padding"
    PYDTNN_STRIDE = "stride"

    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")

    dict_attributes = info[cons.CONST_ATTRIBUTES]
    args = dict()

    if ONNX_COUNT_DILATATIONS in dict_attributes:
        args[PYDTNN_DILATION] = dict_attributes[ONNX_COUNT_DILATATIONS]
    if ONNX_KERNEL_SHAPE in dict_attributes:
        args[PYDTNN_POOL_SHAPE] = dict_attributes[ONNX_KERNEL_SHAPE]
    if ONNX_PADS in dict_attributes:
        args[PYDTNN_PADDING] = cons.pads_from_onnx_to_pydtnn(pads=dict_attributes[ONNX_PADS])
    if ONNX_STRIDES in dict_attributes:
        args[PYDTNN_STRIDE] = dict_attributes[ONNX_STRIDES]

    from pydtnn.layers.average_pool_2d import AveragePool2D
    return AveragePool2D(**args)
# --- END AveragePool --- #

# ========================= #


# ===== #
# = B = #
# ===== #

def BatchNormalization(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")

    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#l-onnx-doc-batchnormalization
    ONNX_EPSILON = "epsilon"
    ONNX_MOMENTUM = "momentum"
    # PyDTNN attributes names from BatchNormalization class.
    PYDTNN_EPSILON = "epsilon"
    PYDTNN_MOMENTUM = "momentum"

    args = dict()
    dict_attributes = info[cons.CONST_ATTRIBUTES]

    if ONNX_EPSILON in dict_attributes:
        args[PYDTNN_EPSILON] = dict_attributes[ONNX_EPSILON]
    if ONNX_MOMENTUM in dict_attributes:
        args[PYDTNN_MOMENTUM] = dict_attributes[ONNX_MOMENTUM]

    from pydtnn.layers.batch_normalization import BatchNormalization
    return BatchNormalization(**args)
# --- END BatchNormalization --- #

# ========================= #


# ===== #
# = C = #
# ===== #

def Concat(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__Concat.html#l-onnx-doc-concat
    ONNX_AXIS = "axis"
    # There are no PyDTNN attributes names from ConcatenationBlock class.

    # TODO: Check if this class is correct
    list_concat_nodes = info[cons.CONST_listS_NODES]

    from pydtnn.layers.concatenation_block import ConcatenationBlock
    return ConcatenationBlock(list_concat_nodes)
# --- END Concat --- #


def Conv(info: dict[str, Any]) -> LayerBase:

    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")

    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__Conv.html#l-onnx-doc-conv
    ONNX_COUNT_DILATATIONS = "dilations"
    ONNX_GROUP = "group"
    ONNX_KERNEL_SHAPE = "kernel_shape"
    ONNX_PADS = "pads"
    ONNX_STRIDES = "strides"
    # PyDTNN attributes names from Conv2D class.
    PYDTNN_DILATION = "dilation"
    PYDTNN_NFILTERS = "nfilters"
    PYDTNN_FILTER_SHAPE = "filter_shape"
    PYDTNN_PADDING = "padding"
    PYDTNN_STRIDE = "stride"

    args = dict()
    dict_attributes = info[cons.CONST_ATTRIBUTES]

    if ONNX_COUNT_DILATATIONS in dict_attributes:
        args[PYDTNN_DILATION] = dict_attributes[ONNX_COUNT_DILATATIONS]
    if ONNX_GROUP in dict_attributes:
        # TODO: Check if this is correct:
        args[PYDTNN_NFILTERS] = dict_attributes[ONNX_GROUP]
    if ONNX_KERNEL_SHAPE in dict_attributes:
        args[PYDTNN_FILTER_SHAPE] = dict_attributes[ONNX_KERNEL_SHAPE]
    if ONNX_PADS in dict_attributes:
        args[PYDTNN_PADDING] = cons.pads_from_onnx_to_pydtnn(pads=dict_attributes[ONNX_PADS])
    if ONNX_STRIDES in dict_attributes:
        args[PYDTNN_STRIDE] = dict_attributes[ONNX_STRIDES]

    # TODO: Look if it's necessary to set the Bias here.

    # TODO: Borrar
    print("CONVOLUCION")
    for k in args.keys():
        print(f"args[{k}]: {type(args[k])} | {args[k]}")

    from pydtnn.layers.conv_2d import Conv2D
    return Conv2D(**args)
# --- END Conv --- #

# ========================= #


# ===== #
# = D = #
# ===== #

def Dropout(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__Dropout.html#l-onnx-doc-dropout
    ONNX_SEED = "seed"  # TODO: Check if the random seed it's important. If it is, check how to set it.
    # PyDTNN attributes names from Dropout class.
    PYDTNN_RATE = "rate"

    args = {}

    # TODO: Check if this is correct.
    # Droput can receive 3 inputs: the previous layer output [Tensor],
    #   the ratio (of random dropout) [Float] and if it's in training mode [bool]
    # Then if it has more than one input and it's not a bool or the previous layer output, it is the ratio.
    _other_inputs = set(enumerate(info[cons.CONST_ALL_INPUTS])) - set(enumerate(info[cons.CONST_INPUTS]))
    other_inputs = [elem[1] for elem in sorted(_other_inputs, key=lambda x: x[0])]

    if len(other_inputs) > 0:
        for k in other_inputs:
            elem = info[cons.CONST_WEIGHTS][other_inputs[k]]
            if not isinstance(elem, bool):
                args[PYDTNN_RATE] = elem
                break

    print(f"==> args: {args}")

    from pydtnn.layers.dropout import Dropout
    return Dropout(**args)
# --- END Dropout --- #

# ========================= #

# ===== #
# = F = #
# ===== #


def Flatten(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    # Source: https://onnx.ai/onnx/operators/onnx__Flatten.html
    # It has one attribute (axis), but there is no equivalence in PyDTNN.
    # ==> In PyDTNN the axis is always 1.
    from pydtnn.layers.flatten import Flatten
    return Flatten()
# --- END Flatten --- #

# ========================= #

# ===== #
# = G = #
# ===== #


def Gemm(info: dict[str, Any]) -> LayerBase:

    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    # Onnx documentation: https://onnx.ai/onnx/operators/onnx__Gemm.html
    ONNX_ALPHA = "alpha"
    ONNX_BETA = "beta"
    ONNX_TRANS_A = "transA"
    ONNX_TRANS_B = "transB"

    # FC' PyDTNN0' implementation:
    #   res = self.model.matmul(x, self.weights)
    #   self.model.tracer.emit_event(PYDTNN_OPS_EVENT, 0)
    #   return res + self.biases if self.use_bias else 0

    # ONNX Gemm implementation:
    #   A’ = transpose(A) if transA else A
    #   B’ = transpose(B) if transB else B
    #   Y = alpha * A’ * B’ + beta * C
    # B: PyDTNN's weights
    # C: PyDTNN's bias

    dict_attributes = info[cons.CONST_ATTRIBUTES]

    alpha = dict_attributes[ONNX_ALPHA] if ONNX_ALPHA in dict_attributes else 1.0
    beta = dict_attributes[ONNX_BETA] if ONNX_BETA in dict_attributes else 1.0
    transA = dict_attributes[ONNX_TRANS_A] if ONNX_TRANS_A in dict_attributes else None
    transB = dict_attributes[ONNX_TRANS_B] if ONNX_TRANS_B in dict_attributes else None

    # TODO: make this programming terrorism into an actual class or classes
    from pydtnn.layers.fc import FC
    pseudo_gemm = FC()

    _other_inputs = set(enumerate(info[cons.CONST_ALL_INPUTS])) - set(enumerate(info[cons.CONST_INPUTS]))
    other_inputs = [elem[1] for elem in sorted(_other_inputs, key=lambda x: x[0])]

    if len(other_inputs) == 1:
        b = info[cons.CONST_WEIGHTS][other_inputs[0]]
        c = None
    else:
        b = info[cons.CONST_WEIGHTS][other_inputs[0]]
        c = info[cons.CONST_WEIGHTS][other_inputs[1]]

    original_fw = pseudo_gemm.forward

    def _weights_initializer(*to_ignore):
        return b.T if transB is not None else b

    def _biases_initializer(*to_ignore):
        return beta * c

    def _mod_forward(x):
        x = alpha * (x.T if transA is not None else x)
        original_fw(x)

    pseudo_gemm.weights_initializer = _weights_initializer  # (lambda *x: b.T if transB is not None else b)
    if c is not None:
        pseudo_gemm.biases_initializer = _biases_initializer
    pseudo_gemm.forward = _mod_forward

    return pseudo_gemm
# --- END Gemm --- #


def GlobalAveragePool(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    # 1.- Onnx documentation: https://onnx.ai/onnx/operators/onnx__GlobalAveragePool.html

    # PyDTNN attributes names from AbstractPool2DLayer class.
    PYDTNN_POOL_SHAPE = "pool_shape"
    PYDTNN_STRIDE = "stride"

    args = dict()

    operations = info[cons.CONST_PREV_LAYERS]
    _input = info[cons.CONST_INPUTS][0]  # It should be a list with only one input

    # TODO: check if this is correct.

    # "This is equivalent to AveragePool with kernel size equal to the spatial dimension of input tensor." [1]
    args[PYDTNN_POOL_SHAPE] = operations[_input].shape
    args[PYDTNN_STRIDE] = 1

    from pydtnn.layers.average_pool_2d import AveragePool2D
    return AveragePool2D(**args)
# --- END GlobalAveragePool --- #


# ========================= #


# ===== #
# = M = #
# ===== #

def MaxPool(info: dict[str, Any]) -> LayerBase:
    print("------")
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")

    # Onnx attributes names from: https://onnx.ai/onnx/operators/onnx__MaxPool.html#l-onnx-doc-maxpool
    ONNX_KERNEL_SHAPE = "kernel_shape"
    ONNX_PADS = "pads"
    ONNX_STRIDES = "strides"
    ONNX_COUNT_DILATATIONS = "dilations"
    # PyDTNN attributes names from AbstractPool2DLayer class.
    PYDTNN_POOL_SHAPE = "pool_shape"
    PYDTNN_PADDING = "padding"
    PYDTNN_STRIDE = "stride"
    PYDTNN_DILATION = "dilation"

    print("DEBUG")  # TODO: ¡¡BORRAR!!

    dict_attributes = info[cons.CONST_ATTRIBUTES]
    args = dict()

    if ONNX_COUNT_DILATATIONS in dict_attributes:
        args[PYDTNN_DILATION] = dict_attributes[ONNX_COUNT_DILATATIONS]
    if ONNX_KERNEL_SHAPE in dict_attributes:
        args[PYDTNN_POOL_SHAPE] = dict_attributes[ONNX_KERNEL_SHAPE]
    if ONNX_PADS in dict_attributes:
        args[PYDTNN_PADDING] = cons.pads_from_onnx_to_pydtnn(pads=dict_attributes[ONNX_PADS])
    if ONNX_STRIDES in dict_attributes:
        args[PYDTNN_STRIDE] = dict_attributes[ONNX_STRIDES]

    for k in args.keys():
        print(f"args[{k}]: {type(args[k])} | {args[k]}")
        a = args[k]

    from pydtnn.layers.max_pool_2d import MaxPool2D
    return MaxPool2D(**args)
# --- END MaxPool --- #


def Mul(info: dict[str, Any]) -> LayerBase:
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")

    # TODO: Move it to a file and do it in the right way.

    from numpy import multiply
    from pydtnn.layers.abstract.block_layer import AbstractBlockLayer

    class _Mul(AbstractBlockLayer):
        def initialize_block_layer(self):
            super().initialize_block_layer()
            assert all([o == self.out_shapes[0] for o in self.out_shapes])
            self.shape = self.out_shapes[0]
        # - END initialize_block_layer - #

        def forward(self, x):
            x = [x] * len(self.paths)
            for i, p in enumerate(self.paths):
                for layer in p:
                    x[i] = layer.forward(x[i])

                if i > 0:
                    # TODO: do it with Cython.
                    x[0] = multiply(x[0], x[i])
            return x[0]
        # - END forward - #

        def backward(self, dy):
            dx = [dy] * len(self.paths)
            for i, p in enumerate(self.paths):
                for layer in reversed(p):
                    dx[i] = layer.backward(dx[i])
                if i > 0:
                    # TODO: do it with Cython adn chekc if it's correct.
                    dx[0] = multiply(dx[0], dx[i])
            return dx[0]
        # - END backward - #
    # -- END _Mul -- #

    return _Mul(info[cons.CONST_listS_NODES])
# --- END Mul --- #

# ========================= #


# ===== #
# = R = #
# ===== #

def Relu(info: dict[str, Any]) -> LayerBase:
    # ONNX info: https://onnx.ai/onnx/operators/onnx__Relu.html
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    from pydtnn.activations.relu import Relu
    return Relu()
# --- END Relu --- #


# ========================= #


# ===== #
# = U = #
# ===== #


def Unsqueeze(info: dict[str, Any]) -> LayerBase:
    # Onnx information: https://onnx.ai/onnx/operators/onnx__Unsqueeze.html
    print(f"attributes: {info[cons.CONST_ATTRIBUTES]}")
    ONNX_AXES = "axes"

    PYDTNN_AXES = "axis"
    dict_attributes = info[cons.CONST_ATTRIBUTES]

    args = {}

    if ONNX_AXES in dict_attributes:
        args[PYDTNN_AXES] = dict_attributes[ONNX_AXES]

    # TODO: Move it to a file and do it in the right way.

    from numpy import expand_dims
    from pydtnn.layers.layer import Layer

    class _Unsqueeze(Layer):

        def __init__(self, shape=(1,), axis=()):
            super().__init__(shape)
            self.axis = axis
        # - END __init__ - #

        def initialize(self, prev_shape, need_dx=False):
            super().initialize(prev_shape, need_dx)
            self.shape = self.shape + self.model.encode_shape(self.model.tensor_format)

        def initialize_block_layer(self):
            super().initialize_block_layer()
            assert all([o == self.out_shapes[0] for o in self.out_shapes])
            self.shape = self.out_shapes[0]
        # - END initialize_block_layer - #

        def forward(self, x):
            return expand_dims(x, axis=self.axis)
    # -- END _Unsqueeze -- #

    return _Unsqueeze(**args)

# --- END Unsqueeze --- #



import pycuda.gpuarray as gpuarray  #type: ignore
from pycuda.driver import Function  #type: ignore

from pydtnn.metrics.metric import Metric
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape


class MetricGPU(Metric[TensorGPU]):
    """
    Extends a Metric class with the attributes and methods required by GPU Metrics.
    """

    def __init__(self, shape: ArrayShape, eps=1e-8):
        super().__init__(shape, eps)
        # NOTE: The following attributes will be initializated later.
        self.grid = None
        self.block = None
    
    def initialize(self) -> None:
        super().initialize()
        self.kernel = self.__init_gpu_kernel__()

        self.grid = self.model.cuda_grid
        self.block = self.model.cuda_block

    def __init_gpu_kernel__(self) -> Function:
        raise NotImplementedError()



from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.backends.gpu.layers.abstract.pool_2d_layer import AbstractPool2DLayerGPU
from pydtnn.libs import libcudnn as cudnn
from pydtnn.backends.gpu.utils.tensor_gpu import TensorGPU
from pydtnn.utils.constants import ArrayShape


class AveragePool2DGPU(AveragePool2D[TensorGPU], AbstractPool2DLayerGPU):

    def initialize(self, prev_shape: ArrayShape, x: TensorGPU) -> None:
        super().initialize(prev_shape, x)
        pool_mode = cudnn.cudnnPoolingMode['CUDNN_POOLING_AVERAGE_COUNT_EXCLUDE_PADDING']
        self.initialize_pool_2d_gpu(prev_shape, x, pool_mode)



import numpy as np
from pydtnn.backends.cpu.utils.transpose_cython import transpose_0231_ijk_cython, transpose_0231_ikj_cython

from pydtnn.utils.best_of import BestOf
from typing import Callable


def transpose_0231_numpy(original: np.ndarray, 
                         transposed: np.ndarray | None = None # type: ignore
                        ) -> np.ndarray:  
    d0, d1, d2, d3 = original.shape
    if transposed is None:
        transposed: np.ndarray = np.empty((d0, d2, d3, d1), original.dtype, order="C")
    transposed[...] = original.transpose((0, 2, 3, 1))
    return transposed


def transpose_0231_ijk_cython_wrapper(original: np.ndarray, 
                                      transposed: np.ndarray | None = None # type: ignore
                                      ) -> np.ndarray:
    d0, d1, d2, d3 = original.shape
    if transposed is None:
        transposed: np.ndarray = np.empty((d0, d2, d3, d1), original.dtype, order="C")
    transpose_0231_ijk_cython(original, transposed)
    return transposed


def transpose_0231_ikj_cython_wrapper(original: np.ndarray, 
                                      transposed: np.ndarray | None = None # type: ignore
                                      ) -> np.ndarray:
    d0, d1, d2, d3 = original.shape
    if transposed is None:
        transposed: np.ndarray = np.empty((d0, d2, d3, d1), original.dtype, order="C")
    transpose_0231_ikj_cython(original, transposed)
    return transposed

# TODO: change typing "Callable, etc."
best_transpose_0231: Callable[[np.ndarray, np.ndarray | None], np.ndarray] = BestOf(
    name="Transpose 0231 methods",
    alternatives=[
        ("ikj_cyt", transpose_0231_ikj_cython_wrapper),
        ("ijk_cyt", transpose_0231_ijk_cython_wrapper),
        ("numpy", transpose_0231_numpy),
    ],
    get_problem_size=lambda *args: args[0].shape,
)



import numpy as np

from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.backends.cpu.metrics.binary_confusion_matrix import BinaryConfusionMatrixCPU
from pydtnn.metrics.f1_score import F1Score
from pydtnn.backends.cpu.utils.div_arrays_set_if_zero import div_arrays_set_if_zero

class F1ScoreCPU(F1Score[np.ndarray], MetricCPU):

    conf_matrix_metric: BinaryConfusionMatrixCPU

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> float:
        true_positives = self.conf_matrix_metric.get_true_positives()
        false_positives = self.conf_matrix_metric.get_false_positives()
        false_negatives = self.conf_matrix_metric.get_false_negatives()

        # 2 * true_positives / (2 * true_positives + false_positives + false_negatives
        f1 = np.multiply(2, true_positives, dtype=np.dtype(float), order="C")
        divider = np.add(f1, false_positives,dtype=np.dtype(float), order="C")
        divider = np.add(divider, false_negatives, dtype=np.dtype(float), order="C")

        div_arrays_set_if_zero(f1,  divider, default_value=0.0)
        return float(np.average(f1))



# _______________________________________________________________________________________________________________
# In this file must be implemented only the translation of PyTorch Activations layers to its PyDTNN equivalent.
# _______________________________________________________________________________________________________________

# Typing related (or non important) imports
from typing import Dict, Any


# Functionality imports
from pydtnn.activations.arctanh import Arctanh as _Arctanh
from pydtnn.activations.log import Log as _Log
from pydtnn.activations.relu import Relu as _Relu
from pydtnn.activations.relu6 import Relu6 as _Relu6
from pydtnn.activations.leaky_relu import LeakyRelu as _LeakyRelu
from pydtnn.activations.sigmoid import Sigmoid as _Sigmoid
from pydtnn.activations.softmax import Softmax as _Softmax
from pydtnn.activations.tanh import Tanh as _Tanh
import pydtnn.converters.pytorch2pydtnn.common as cm
# ------------------- #


def Arctanh(args: Dict[str, Any]) -> _Arctanh:
    # NOTE: There is no equivalent in PyTorch
    not_used = args
    return _Arctanh()


def LogSigmoid(args: Dict[str, Any]) -> _Log:
    # https://pytorch.org/docs/stable/generated/torch.nn.LogSigmoid.html#torch.nn.LogSigmoid
    not_used = args
    return _Log()


def ReLU(args: Dict[str, Any]) -> _Relu:
    # https://pytorch.org/docs/stable/generated/torch.nn.ReLU.html#torch.nn.ReLU
    # Not used Pytorch's parameters: inplace.
    not_used = args
    return _Relu()


def ReLU6(args: Dict[str, Any]) -> _Relu:
    # https://pytorch.org/docs/stable/generated/torch.nn.ReLU.html#torch.nn.ReLU
    # Not used Pytorch's parameters: inplace.
    not_used = args

    # NOTE: max_val. A interal PyTorch variable that seems to set the cap.

    return _Relu6()


def LeakyReLU(args: Dict[str, Any]) -> _Relu:
    # https://pytorch.org/docs/stable/generated/torch.nn.ReLU.html#torch.nn.ReLU
    # Not used Pytorch's parameters: inplace.
    NEGATIVE_SLOPE = "negative_slope"
    torch_dict_keys = [NEGATIVE_SLOPE]
    pydtnn_dict_keys = [NEGATIVE_SLOPE]

    layer_args = cm.prepare_pydtnn_arguments(arguments=args[cm.ARGUMENTS], torch_dict_keys=torch_dict_keys, pydtnn_dict_keys=pydtnn_dict_keys)

    return _LeakyRelu(**layer_args)


def Sigmoid(args: Dict[str, Any]) -> _Sigmoid:
    # https://pytorch.org/docs/stable/generated/torch.nn.Sigmoid.html#torch.nn.Sigmoid
    not_used = args
    return _Sigmoid()


def Softmax(args: Dict[str, Any]) -> _Softmax:
    # https://pytorch.org/docs/stable/generated/torch.nn.Softmax.html#torch.nn.Softmax
    # Not used Pytorch's parameters: dim.
    not_used = args
    return _Softmax()


def Tanh(args: Dict[str, Any]) -> _Tanh:
    # https://pytorch.org/docs/stable/generated/torch.nn.Tanh.html#torch.nn.Tanh
    not_used = args
    return _Tanh()



from pydtnn.activations.relu import Relu
from pydtnn.utils.constants import ArrayShape, Array

# NOTE -> "CappedRelu": https://www.cs.utoronto.ca/~kriz/conv-cifar10-aug2010.pdf


class Relu6[T: Array](Relu[T]):
    # NOTE: This is a ReLU6 *iif* cap is 6, but it's more interesting a implementation where the user have the freedom to choose their cap.
    def __init__(self, shape: ArrayShape = (1,), cap: float = 6.0):
        super().__init__(shape)
        self.cap: float = cap



from collections.abc import Sequence

from pydtnn.activations.softmax import Softmax
from pydtnn.activations.relu import Relu
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.concatenation_block import ConcatenationBlock
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase
from pydtnn.utils.initializers import he_uniform
from pydtnn.layers.max_pool_2d import MaxPool2D


def inceptionv3(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    _(Input(shape=input_shape))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), stride=2, weights_initializer=he_uniform))
    _(Conv2D(nfilters=32, filter_shape=(3, 3), weights_initializer=he_uniform))
    _(Conv2D(nfilters=64, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform))
    _(MaxPool2D(pool_shape=(3, 3), stride=2))
    _(Conv2D(nfilters=80, filter_shape=(1, 1), weights_initializer=he_uniform))
    _(Conv2D(nfilters=192, filter_shape=(3, 3), weights_initializer=he_uniform))
    _(MaxPool2D(pool_shape=(3, 3), stride=2))

    inception_blocks = [[64, 48, 64, 64, 96, 32],
                        [64, 48, 64, 64, 96, 64],
                        [64, 48, 64, 64, 96, 64]]

    for n1x1, n5x5red, n5x5, n3x3red, n3x3, pool_planes in inception_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n1x1, filter_shape=(1, 1), weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n5x5red, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n5x5, filter_shape=(5, 5), padding=2, weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n3x3red, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform)
             ],
            [AveragePool2D(pool_shape=(3, 3), stride=1, padding=1),
             Conv2D(nfilters=pool_planes, filter_shape=(1, 1), weights_initializer=he_uniform)
             ]))

    inception_blocks = [[384, 64, 96]]

    for n1x1, n3x3red, n3x3 in inception_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n1x1, filter_shape=(3, 3), stride=2, padding=0, weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n3x3red, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3, filter_shape=(3, 3), weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3, filter_shape=(3, 3), stride=2, padding=1, weights_initializer=he_uniform)
             ],
            [MaxPool2D(pool_shape=(3, 3), stride=2, padding=0)
             ]))

    inception_blocks = [[192, 128],
                        [192, 160],
                        [192, 160],
                        [192, 192]]

    for n1x1, n1x7 in inception_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n1x1, filter_shape=(1, 1), weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n1x7, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x7, filter_shape=(1, 7), padding=(3, 0), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1, filter_shape=(7, 1), padding=(0, 3), weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n1x7, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x7, filter_shape=(7, 1), padding=(0, 3), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x7, filter_shape=(1, 7), padding=(3, 0), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x7, filter_shape=(7, 1), padding=(0, 3), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1, filter_shape=(1, 7), padding=(3, 0), weights_initializer=he_uniform),
             ],
            [AveragePool2D(pool_shape=(3, 3), stride=1, padding=1),
             Conv2D(nfilters=n1x1, filter_shape=(1, 1), weights_initializer=he_uniform)
             ]))

    inception_blocks = [[192, 320]]

    for n1x1, n3x3 in inception_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n1x1, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n3x3, filter_shape=(3, 3), stride=2, weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n1x1, filter_shape=(1, 1), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1, filter_shape=(1, 7), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1, filter_shape=(7, 1), padding=1, weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1, filter_shape=(3, 3), stride=2, weights_initializer=he_uniform)
             ],
            [MaxPool2D(pool_shape=(3, 3), stride=2)
             ]))

    inception_blocks = [[320, 384, 448, 192],
                        [320, 384, 448, 192]]

    for n1x1b0, n1x1b1, n1x1b2, n1x1b3 in inception_blocks:
        _(ConcatenationBlock(
            [Conv2D(nfilters=n1x1b0, filter_shape=(1, 1), weights_initializer=he_uniform)
             ],
            [Conv2D(nfilters=n1x1b1, filter_shape=(1, 1), weights_initializer=he_uniform),
             ConcatenationBlock(
                 [Conv2D(nfilters=n1x1b1, filter_shape=(1, 3), padding=(0, 1), weights_initializer=he_uniform)],
                 [Conv2D(nfilters=n1x1b1, filter_shape=(3, 1), padding=(1, 0), weights_initializer=he_uniform)])
             ],
            [Conv2D(nfilters=n1x1b2, filter_shape=(1, 1), weights_initializer=he_uniform),
             Conv2D(nfilters=n1x1b1, filter_shape=(3, 3), padding=1, weights_initializer=he_uniform),
             ConcatenationBlock(
                 [Conv2D(nfilters=n1x1b1, filter_shape=(1, 3), padding=(0, 1), weights_initializer=he_uniform)],
                 [Conv2D(nfilters=n1x1b1, filter_shape=(3, 1), padding=(1, 0), weights_initializer=he_uniform)])
             ],
            [AveragePool2D(pool_shape=(3, 3)),
             Conv2D(nfilters=n1x1b3, filter_shape=(1, 1), padding=1, weights_initializer=he_uniform)
             ]))

    _(AveragePool2D(pool_shape=(8, 8), stride=1))  # Global average pooling 2D
    _(Flatten())
    _(FC(shape=(1024,)))
    _(BatchNormalization())
    _(Relu())
    _(FC(shape=output_shape, activation=Softmax))

    return model



from typing import TYPE_CHECKING

from pydtnn.schedulers.scheduler import Scheduler

if TYPE_CHECKING:
    from pydtnn.model import Model


class SchedulerWithLossOrMetric(Scheduler):
    """
    Scheduler with metric base class
    """

    def __init__(self, loss_or_metric: str, verbose: bool):
        # NOTE: loss_or_metric default value is "val_accuracy" in Parser.
        super().__init__(verbose)
        self.is_val_metric: bool = "val_" == loss_or_metric[:4]
        self.loss_or_metric = loss_or_metric[4:] if self.is_val_metric else loss_or_metric

    def _get_idx(self):
        try:
            return self.model.loss_and_metrics.index(self.loss_or_metric)
        except ValueError as e:
            raise ValueError("{self}: loss or metric '{self.loss_or_metric}' not found in current model!") from e

    @classmethod
    def from_model(cls, model: "Model") -> "WarmUpLRScheduler":
        return 


from collections.abc import Sequence

from pydtnn.activations.relu import Relu
from pydtnn.activations.softmax import Softmax
from pydtnn.layers.average_pool_2d import AveragePool2D
from pydtnn.layers.batch_normalization import BatchNormalization
from pydtnn.layers.conv_2d import Conv2D
from pydtnn.layers.fc import FC
from pydtnn.layers.flatten import Flatten
from pydtnn.layers.input import Input
from pydtnn.layer_base import LayerBase


def mobilenetv1(input_shape: Sequence[int], output_shape: Sequence[int]) -> Sequence[LayerBase]:
    model = list[LayerBase]()
    _ = model.append

    first_filters = 32
    _(Input(shape=input_shape))
    _(Conv2D(nfilters=first_filters, filter_shape=(3, 3), grouping=Conv2D.Grouping.STANDARD, padding=1, stride=2, activation=Relu, use_bias=False))

    layout = [[64, 1], [128, 2], [256, 2], [512, 6], [1024, 2]]
    for n_filt, reps in layout:
        for r in range(reps):
            stride = 2 if reps > 1 and r == 0 else 1
            _(Conv2D(nfilters=first_filters, filter_shape=(3, 3), grouping=Conv2D.Grouping.DEPTHWISE, padding=1, stride=stride, use_bias=False))
            _(BatchNormalization())
            _(Relu())
            _(Conv2D(nfilters=n_filt, filter_shape=(1, 1), grouping=Conv2D.Grouping.POINTWISE, use_bias=False))
            _(BatchNormalization())
            _(Relu())
            first_filters = n_filt

    _(AveragePool2D(pool_shape=(1, 1)))
    _(Flatten())
    _(FC(shape=(1024,)))
    _(FC(shape=output_shape, activation=Softmax))

    return model



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.datasets.dataset import Dataset
from pydtnn.utils import random


if TYPE_CHECKING:
    from pydtnn.model import Model

TRAIN_NSAMPLES = 10000
TEST_NSAMPLES = 2000
INPUT_SHAPE = (1,)
OUTPUT_SHAPE = (1,)


class MaskLang(Dataset):
    """
    Masked Language Model Dataset

    NOTE: Source unclear
    TODO: Load original dataset

    Source (SHA1): ???

    Normalize (z-score):
    offset: ???
    scale:  ???
    """

    def __init__(self, model: "Model", preprocess=True, embedl=512, max_sentence=512, split_token="<translation>", force_test_as_validation=False, debug=False):
        super().__init__(model, TRAIN_NSAMPLES, TEST_NSAMPLES, INPUT_SHAPE, OUTPUT_SHAPE, force_test_as_validation=force_test_as_validation, debug=debug)

        self.do_preprocess = preprocess
        self.split_token = split_token
        self.max_sentence = max_sentence
        self.embedl = embedl
        self.train_path = self.model.dataset_path
        self.test_path = self.model.dataset_path
        self.test_as_validation = force_test_as_validation
        self.dtype = self.model.dtype
        self.model = model
        self.lang = self.model.dataset_lang

    def _init_actual_data(self):
        # Actual
        self.load_data()
        self.make_train_val_partitions()
        if not self.do_preprocess:
            self._actual_data_generator = self._actual_data_generator_normal
        else:
            self.preprocess(size=197988)
            self.make_train_val_partitions()
            self._actual_data_generator = self._actual_data_generator_preprocess

        # Synthetic
        # self.train_val_nsamples = 10000
        # self.train_nsamples = None
        # self.make_train_val_partitions()
        # self.src_embeddings = random.random((self.train_val_nsamples, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)
        # self.tgt_embeddings = random.random((self.train_val_nsamples, 1, self.max_sentence, self.embedl)).astype(dtype=self.dtype)

    def load_data(self):
        self.dictionary = self.get_dictionary(self.lang)
        self.mask = self.dictionary("Mask")[0]
        self.mask = np.zeros(self.mask.vector.shape, dtype=self.dtype)
        file = open(self.train_path, "r")
        self.lines = file.readlines()
        file.close()
        self.lines = [line.replace("\n", "") for line in self.lines]

        self.train_val_nsamples = len(self.lines)
        print(self.train_val_nsamples)
        self.train_nsamples = None

    def get_dictionary(self, language):
        import spacy
        table = {
            "en": "en_core_web_md",
            "de": "de_core_news_md"
        }
        if language in table:
            language = table[language]
        return spacy.load(language)

    def make_train_val_partitions(self):
        val_split = self.model.validation_split
        if self.train_nsamples == None:
            s = np.arange(self.train_val_nsamples)
            if self.model.augment_shuffle:
                random.shuffle(s)
            self.train_nsamples = int(self.train_val_nsamples * (1-val_split) // 1)
            self.train_indices = s[:self.train_nsamples]
            self.val_indices = s[self.train_nsamples:]
            self.val_nsamples = len(self.val_indices)
            self.test_nsamples = self.val_nsamples

    def _actual_data_generator_normal(self, part):
        batch_size = self.model.batch_size
        rank = self.model.rank

        for i in range(self.train_val_nsamples // batch_size):
            window = (i * batch_size + rank * batch_size, i * batch_size + (rank + 1) * batch_size)
            src_embeddings = np.zeros((batch_size, 1, self.max_sentence, self.embedl), dtype=self.dtype)
            tgt_embeddings = np.zeros((batch_size, 1, self.max_sentence, self.embedl), dtype=self.dtype)
            for i, doc in enumerate(self.dictionary.pipe(self.lines[window[0]:window[1]])):
                mask = np.random.randint(0, len(doc))
                for j, word in enumerate(doc):
                    if j > self.max_sentence:
                        break
                    if j == mask:
                        src_embeddings[i, 0, j] = self.mask
                    else:
                        src_embeddings[i, 0, j] = word.vector
                    tgt_embeddings[i, 0, j] = word.vector
            x = src_embeddings
            y = tgt_embeddings
            yield x, y

    def _actual_data_generator_preprocess(self):
        batch_size = self.model.batch_size
        rank = self.model.rank

        for i in range(self.train_val_nsamples // batch_size):
            window = (i * batch_size + rank * batch_size, i * batch_size + (rank + 1) * batch_size)
            x = self.src_embeddings[window[0]:window[1]]
            y = self.tgt_embeddings[window[0]:window[1]]
            yield x, y

    def _synthetic_data_generator(self):
        batch_size = self.model.batch_size
        rank = self.model.rank

        for i in range(self.train_val_nsamples // batch_size):
            window = (i * batch_size + rank * batch_size, i * batch_size + (rank + 1) * batch_size)
            x = self.src_embeddings[window[0]:window[1]]
            y = self.tgt_embeddings[window[0]:window[1]]
            yield x, y

    # === Preprocess ===
    def preprocess(self, size=None):
        if size is None:
            size = len(self.lines)
        self.train_val_nsamples = size
        self.src_embeddings = np.zeros((size, 1, self.max_sentence, self.embedl), dtype=self.dtype)
        self.tgt_embeddings = np.zeros((size, 1, self.max_sentence, self.embedl), dtype=self.dtype)
        for i, doc in enumerate(self.dictionary.pipe(self.lines[0:size])):
            mask = np.random.randint(0, len(doc))
            # self.src_embeddings[i,0,0:len(doc)] = doc
            # self.tgt_embeddings[i,0,]
            # self.src_embeddings[i,0,mask] = self.mask
            for j, word in enumerate(doc):
                if j > self.max_sentence:
                    break
                if j == mask:
                    self.src_embeddings[i, 0, j] = self.mask
                else:
                    self.src_embeddings[i, 0, j] = word.vector
    # === Preprocess ===



import numpy as _np
type _npDT = _np.int8 | _np.float32 | _np.float64
type _npDT_4Dims[T] = _np.ndarray[tuple[int, int, int, int], T]
type _npDT_3Dims[T] = _np.ndarray[tuple[int, int, int], T]
type _npDT_2Dims[T] = _np.ndarray[tuple[int, int], T]
type _npDT_1Dims[T] = _np.ndarray[tuple[int], T]


def bn_training_fwd_cython[T: _npDT](x: _npDT_2Dims[T],
                                     y: _npDT_2Dims[T],
                                     xn: _npDT_2Dims[T],
                                     std: _npDT_1Dims[T],
                                     gamma: _npDT_1Dims[T],
                                     beta: _npDT_1Dims[T],
                                     mean: _npDT_1Dims[T],
                                     var: _npDT_1Dims[T],
                                     eps: float) -> None:
    """
    Args:
        x (npDT_2Dims): The input.
        y (npDT_2Dims): 2-dimensional ndarray where the output is stored.
        xn (npDT_2Dims): 2-dimensional ndarray where the mean of x is stored.
        std (npDT_1Dims): 2-dimensional ndarray where the standard deviation of x is stored.
        gamma (npDT_1Dims): batch normalization gamma.
        beta (npDT_1Dims): batch normalization beta.
        mean (npDT_1Dims): input means.
        var (npDT_1Dims): input variances.
        eps (float): epsilon.

    Returns:
        out: Nothing. The output is stored in \"y\", \"xn\" and \"std\".

    Note:
        It's never used.
    """
    ...
# ---


def bn_training_bwd_cython[T: _npDT](dx: _npDT_2Dims[T],
                                     dy: _npDT_2Dims[T],
                                     xn: _npDT_2Dims[T],
                                     std: _npDT_1Dims[T],
                                     gamma: _npDT_1Dims[T],
                                     dgamma: _npDT_1Dims[T],
                                     dbeta: _npDT_1Dims[T]) -> None:
    """
    Args:
        dx (npDT_2Dims): The 2 dimensional array that contains the gradient of the input forward's (that is the output).
        dy (npDT_2Dims): The 2 dimensional array that contains the gradient of the backward's input.
        xn (npDT_2Dims): The 2 dimensional array that contains the normalized input's value.
        std (npDT_1Dims): The 1 dimensions ndarray that stores the standard deviation
        gamma (npDT_1Dims): The input's 1 dimensions thtat stores the gamma's values
        dgamma (npDT_1Dims): The input's 1 dimensions ndarray the gradient of the gamma's values
        dbeta (npDT_1Dims): The input's 1 dimensions ndarray the  gradient of the beta's values

    Returns:
        Nothing. The output will be stored in "dx".
    """
    ...



"""
PyDTNN convGemm module
"""

import ctypes
import platform
import weakref

import numpy as np
from pydtnn.backends.cpu.utils.im2col_nchw_cython import im2col_nchw_cython

from pydtnn.utils import load_library

try:
    load_library("convGemm")
    is_conv_gemm_available = True
except Exception:
    is_conv_gemm_available = False


class ConvGemm:
    """
    Exposes the libconvGemm functions following the PyDTNN conventions.

    Methods
    -------
    conv_gemm(weights, x, out, vpadding, hpadding, vstride, hstride,
              vdilation, hdilation, biases)
        Calls the appropriate convGemm function from libconvGemm.so to perform a
        matrix matrix multiplication with an implicit im2col.

    Examples
    --------
    See __usage_example__() method for an example of use. This example can be
    run with: 'python conv_gemm.py'

    Tests
    -----
    To perform the tests, run the following command from the current directory:
        python -m unittest tests.ConvGemmTestCase

    (see tests/conv_gemm.py for more instructions on testing)
    """

    lib_cg = None  # will link to the libconvGemm.so library

    def __init__(self, dtype: np.dtype = np.dtype(np.float32), debug: bool = False, parent_layer=None):
        """
        Loads the libconvGemm.so library and creates the required auxiliary matrices ac_pack and bc_pack.

        Parameters
        ----------
        dtype : data type
            The element data type being used on all the matrices.
        debug : boolean
            Whether to print debug information or not.
        parent_layer: layer
            The layer that is using it (for tracing purposes).
        """
        self.dtype = dtype
        if ConvGemm.lib_cg is None:
            ConvGemm.lib_cg = load_library("convGemm")

        # Declare ac_pack and bc_pack and allocate space for them
        self.ac_pack = ctypes.POINTER(ctypes.c_float)()
        self.bc_pack = ctypes.POINTER(ctypes.c_float)()
        self.lib_cg.alloc_pack_buffs.restype = ctypes.c_int
        result = self.lib_cg.alloc_pack_buffs(ctypes.byref(self.ac_pack), ctypes.byref(self.bc_pack))
        if result == 1:
            raise MemoryError("Could not allocate space for ac_pack or bc_pack!")
        # Debug
        self.debug = debug
        # Parent layer
        if parent_layer is not None:
            self.get_parent_layer = weakref.ref(parent_layer)
        # Choose the appropriate convGemm function depending on the architecture and the data type being used
        if self.dtype == np.float32:
            self.x_conv_gemm_nhwc = self.lib_cg.sconvGemmNHWC
            self.x_deconv_gemm_nhwc = self.lib_cg.sconvGemmNHWC_back
            self.x_conv_gemm_nchw = self.lib_cg.sconvGemmNCHW
            self.x_deconv_gemm_nchw = self.lib_cg.sconvGemmNCHW_back
        else:
            raise TypeError(f"Type '{str(self.dtype)}' not supported by this version of libconvGemm!")

    def __del__(self):
        """Free the allocated matrices"""
        try:
            __free__(self.ac_pack)
            __free__(self.bc_pack)
        except AttributeError:
            pass

    def conv_gemm_nchw(self, weights: np.ndarray, x: np.ndarray,
                       # res originaly was called "biases"
                       out: np.ndarray | None = None,  # type: ignore
                       vpadding=0, hpadding=0, vstride=1, hstride=1,
                       vdilation=1, hdilation=1,
                       # biases originaly was called "biases_vector"
                       biases: np.ndarray | None = None,   # type: ignore
                       trans=False,
                       bn_running_mean: np.ndarray | None = None,  # type: ignore
                       bn_inv_std: np.ndarray | None = None,  # type: ignore
                       bn_gamma: np.ndarray | None = None,   # type: ignore
                       bn_beta: np.ndarray | None = None,   # type: ignore
                       relu=False):
        """
        Calls the appropriate convGemm function from libconvGemm.so to perform a
        matrix matrix multiplication with an implicit im2col.

        The matrix matrix product is in the form C = A * B, where:
            + A is the weights matrix,
            + B is the im2col(x) matrix, and
            + C is the out matrix.

        If the out vector is supplied, the xapplyBias function of the libconvGemm library will be called. This
        function sums each element of the out vector to all the elements in the corresponding output channel.

        Parameters
        ----------
        weights : array_like
            The weights matrix (kn x c x kh x kw).
        x : array_like
            The layers matrix (b x c x h x w).
        out : array_like
            An optional out matrix (kn x b*ho*wo). If provided, can be overwritten.
        vpadding : int
            The vertical padding to be applied to the x matrix.
        hpadding : int
            The horizontal padding to be applied to the x matrix.
        vstride : int
            The vertical stride.
        hstride : int
            The horizontal stride.
        vdilation : int
            The vertical dilation.
        hdilation : int
            The horizontal dilation.
        biases: array_like
            The out that have to be summed to all the elements in each output channel.
        trans: bool
            Perform the im2col(x) if False, or the im2colT(x) if True.

        Returns
        -------
        array_like
            The result of weights * im2col(x)
        """

        # Get matrices dimensions
        b, c, h, w = x.shape
        if not trans:
            kn, ck, kh, kw = weights.shape
            ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
            wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1
            if out is None:
                out = np.zeros((b, kn, ho, wo), weights.dtype, order="C")
            else:
                out = out[:b, :]
                bb, knb, hob, wob = out.shape
                assert bb == b, "Batch size of the out must be the same as in the input!"
                assert knb == kn, "Number of filters in out must be the same as in the filter tensor!"
                assert hob == ho, "Biases image height must be the same as the output image height!"
                assert wob == wo, "Biases image width must be the same as the output image width!"
        else:
            assert out is not None, "If using the transposed convGemm, the out matrix must be supplied"
            kn, ck, kh, kw = out.shape
            bw, knw, ho, wo = weights.shape
            assert kn == knw, "Number of filters must be the same!"
            assert b == bw, "Batch size must be the same!"
        assert ck == c, "Number of channels in weights and x should be the same!"

        out: np.ndarray
        biases: np.ndarray
        bn_running_mean: np.ndarray
        bn_inv_std: np.ndarray
        bn_gamma: np.ndarray
        bn_beta: np.ndarray

        # Check that dtype is the same on all the matrices
        assert weights.dtype == x.dtype == out.dtype, \
            "All the matrices must have the same type of data!"
        assert weights.dtype == self.dtype, \
            "The input matrices must have the same type of data as the one specified when " \
            "this class was instantiated!"

        # Call the appropriate convGemm function from libconvGemm
        self.x_conv_gemm_nchw(ctypes.c_char(b'Y' if trans else b'N'),
                              ctypes.c_int(b), ctypes.c_int(c), ctypes.c_int(h), ctypes.c_int(w),
                              ctypes.c_int(kn), ctypes.c_int(kh), ctypes.c_int(kw),
                              ctypes.c_int(vpadding), ctypes.c_int(hpadding),
                              ctypes.c_int(vstride), ctypes.c_int(hstride),
                              ctypes.c_int(vdilation), ctypes.c_int(hdilation),
                              ctypes.c_void_p(weights.ctypes.data),
                              ctypes.c_void_p(x.ctypes.data),
                              ctypes.c_void_p(out.ctypes.data),
                              ctypes.c_void_p(None if biases is None else biases.ctypes.data),
                              ctypes.c_void_p(None if bn_running_mean is None else bn_running_mean.ctypes.data),
                              ctypes.c_void_p(None if bn_inv_std is None else bn_inv_std.ctypes.data),
                              ctypes.c_void_p(None if bn_gamma is None else bn_gamma.ctypes.data),
                              ctypes.c_void_p(None if bn_beta is None else bn_beta.ctypes.data), ctypes.c_bool(relu),
                              self.ac_pack, self.bc_pack)

        return out

    # TODO: Check for what is out used inside "x_conv_gemm_nhwc" (and set better varible names).
    def conv_gemm_nhwc(self, weights: np.ndarray, x: np.ndarray,
                       out: np.ndarray | None = None,  # type: ignore
                       vpadding=0, hpadding=0, vstride=1, hstride=1,
                       vdilation=1, hdilation=1,
                       biases: np.ndarray | None = None,   # type: ignore
                       trans=False,
                       bn_running_mean: np.ndarray | None = None,  # type: ignore
                       bn_inv_std: np.ndarray | None = None,  # type: ignore
                       bn_gamma: np.ndarray | None = None,   # type: ignore
                       bn_beta: np.ndarray | None = None,   # type: ignore
                       relu=False):

        # Get matrices dimensions
        b, h, w, c = x.shape
        if not trans:
            ck, kh, kw, kn = weights.shape
            ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
            wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1
            if out is None:
                out = np.zeros((b, ho, wo, kn), weights.dtype, order="C")
            else:
                out = out[:b, :]
                bb, hob, wob, knb = out.shape
                assert bb == b, "Batch size of the out must be the same as in the input!"
                assert hob == ho, "Biases image height must be the same as the output image height!"
                assert wob == wo, "Biases image width must be the same as the output image width!"
                assert knb == kn, "Number of filters in out must be the same as in the filter tensor!"
        else:
            assert out is not None, "If using the transposed convGemm, the output matrix must be supplied"
            ck, kh, kw, kn = out.shape
            bw, ho, wo, knw = weights.shape
            assert kn == knw, "Number of filters must be the same!"
            assert b == bw, "Batch size must be the same!"
        assert ck == c, "Number of channels in weights and x should be the same!"

        out: np.ndarray
        biases: np.ndarray
        bn_running_mean: np.ndarray
        bn_inv_std: np.ndarray
        bn_gamma: np.ndarray
        bn_beta: np.ndarray

        # Check that dtype is the same on all the matrices
        assert weights.dtype == x.dtype == out.dtype, \
            "All the matrices must have the same type of data!"
        assert weights.dtype == self.dtype, \
            "The input matrices must have the same type of data as the one specified when " \
            "this class was instantiated!"

        # Call the appropriate convGemm function from libconvGemm
        self.x_conv_gemm_nhwc(ctypes.c_char(b'Y' if trans else b'N'),
                              ctypes.c_int(b), ctypes.c_int(h), ctypes.c_int(w), ctypes.c_int(c),
                              ctypes.c_int(kn), ctypes.c_int(kh), ctypes.c_int(kw),
                              ctypes.c_int(vpadding), ctypes.c_int(hpadding),
                              ctypes.c_int(vstride), ctypes.c_int(hstride),
                              ctypes.c_int(vdilation), ctypes.c_int(hdilation),
                              ctypes.c_void_p(weights.ctypes.data),
                              ctypes.c_void_p(x.ctypes.data),
                              ctypes.c_void_p(out.ctypes.data),
                              ctypes.c_void_p(None if biases is None else biases.ctypes.data),
                              ctypes.c_void_p(None if bn_running_mean is None else bn_running_mean.ctypes.data),
                              ctypes.c_void_p(None if bn_inv_std is None else bn_inv_std.ctypes.data),
                              ctypes.c_void_p(None if bn_gamma is None else bn_gamma.ctypes.data),
                              ctypes.c_void_p(None if bn_beta is None else bn_beta.ctypes.data), ctypes.c_bool(relu),
                              self.ac_pack, self.bc_pack)

        return out

    def deconv_gemm_nchw(self, weights: np.ndarray,
                         dy: np.ndarray,
                         dx: np.ndarray,
                         vpadding=0, hpadding=0,
                         vstride=1, hstride=1,
                         vdilation=1, hdilation=1):
        """
        Calls the appropriate deconv_gemm function from libconvGemm.so to perform
        an inplace matrix matrix multiplication and deconvolution:

            dx = col2im(weights_2D_T * dy_2D),

        where:
          * weights_2D_T is the weights matrix reshaped to 2D and transposed (c·kh·kw x kn),
          * dy_2D is the dy matrix transposed_1023 and reshaped to 2D (kn x b·ho·wo).

        Parameters
        ----------
        weights : array_like
            The weights matrix (kn x c x kh x kw).
        dy : array_like
            The dy matrix (b x kn x ho x wo).
        dx : array_like
            An empty dx matrix (b x c x h x w) that will be overwritten with col2im(* weights_2D_T * dy_2D).
        vpadding : int
            The vertical padding to be applied to the x matrix.
        hpadding : int
            The horizontal padding to be applied to the x matrix.
        vstride : int
            The vertical stride.
        hstride : int
            The horizontal stride.
        vdilation : int
            The vertical dilation.
        hdilation : int
            The horizontal dilation.

        Returns
        -------
        array_like
            The dx matrix.
        """

        # Get matrices dimensions
        kn, ck, kh, kw = weights.shape
        b2, kn2, ho, wo = dy.shape
        b, c, h, w = dx.shape
        assert kn == kn2, "Number of filters outputs in weights and dy should be the same!"
        assert b == b2, "Different batch size!"
        assert ck == c, "Number of channels in weights and x should be the same!"

        self.x_deconv_gemm_nchw(ctypes.c_int(b), ctypes.c_int(c), ctypes.c_int(h), ctypes.c_int(w),
                                ctypes.c_int(kn), ctypes.c_int(kh), ctypes.c_int(kw),
                                ctypes.c_int(vstride), ctypes.c_int(hstride),
                                ctypes.c_int(vpadding), ctypes.c_int(hpadding),
                                ctypes.c_int(vdilation), ctypes.c_int(hdilation),
                                ctypes.c_void_p(weights.ctypes.data),
                                ctypes.c_void_p(dy.ctypes.data),
                                ctypes.c_void_p(dx.ctypes.data),
                                self.ac_pack, self.bc_pack)

        return dx

    def deconv_gemm_nhwc(self, weights: np.ndarray,
                         dy: np.ndarray,
                         dx: np.ndarray,
                         vpadding=0, hpadding=0,
                         vstride=1, hstride=1,
                         vdilation=1, hdilation=1):

        ck, kh, kw, kn = weights.shape
        b2, ho, wo, kn2 = dy.shape
        b, h, w, c = dx.shape
        assert kn == kn2, "Number of filters outputs in weights and dy should be the same!"
        assert b == b2, "Different batch size!"
        assert ck == c, "Number of channels in weights and x should be the same!"

        self.x_deconv_gemm_nhwc(ctypes.c_int(b), ctypes.c_int(h), ctypes.c_int(w), ctypes.c_int(c),
                                ctypes.c_int(kn), ctypes.c_int(kh), ctypes.c_int(kw),
                                ctypes.c_int(vstride), ctypes.c_int(hstride),
                                ctypes.c_int(vpadding), ctypes.c_int(hpadding),
                                ctypes.c_int(vdilation), ctypes.c_int(hdilation),
                                ctypes.c_void_p(weights.ctypes.data),
                                ctypes.c_void_p(dy.ctypes.data), ctypes.c_void_p(dx.ctypes.data),
                                self.ac_pack, self.bc_pack)

        return dx


def __free__(pack):
    def find_msvcr():
        import re
        import sys
        exec_bytes = open(sys.executable, "rb").read()
        match = re.search("msvcr([0-9]+|t).dll", str(exec_bytes), re.IGNORECASE)
        return match.group(0)

    if platform.system() == 'Windows':
        libc = ctypes.cdll.LoadLibrary(find_msvcr())
    elif platform.system() == 'Linux':
        libc = ctypes.cdll.LoadLibrary('libc.so.6')
    elif platform.system == 'Darwin':
        libc = ctypes.cdll.LoadLibrary('libc.dylib')
    else:
        raise AssertionError("Don't know how to get to libc for a '{}' system".format(platform.system()))
    assert isinstance(pack, object)
    libc.free(pack)


def time_it_func(x: np.ndarray, w_c: np.ndarray, out: np.ndarray,
                 b: int, kn: int,
                 ho: int, wo: int, kh: int, kw: int,
                 vpadding: int, hpadding: int, vstride: int, hstride: int,
                 vdilation: int, hdilation: int,
                 ) -> int | float:

    res = np.zeros(((x.shape[0] * ho * wo), (x.shape[-1] * kh * kw)), dtype=x.dtype)
    im2col_nchw_cython(x, res,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride,
                       vdilation, hdilation)
    res = res @ w_c
    res += out.reshape(b * ho * wo, kn)
    return res


def __usage_example__():
    # Imports for this usage example (not required otherwise)
    from timeit import timeit
    # Default parameters (1st layer AlexNet for Cifar10)
    b = 64  # Batch size
    c = 3  # Channels per layer
    h = 32  # Layers height
    w = 32  # Layers width
    kn = 64  # Number of filters
    kh = 3  # Filters weights height
    kw = 3  # Filters weights width
    vpadding = 1  # Vertical padding
    hpadding = 1  # Horizontal padding
    vstride = 2  # Vertical stride
    hstride = 2  # Horizontal stride
    vdilation = 1  # Vertical dilation
    hdilation = 1  # Horizontal dilation
    # Create weights, x, and out matrices from previous parameters. If no out
    # matrix is provided, a proper one filled with zeros will be automatically
    # created.
    weights = np.zeros((kn, c, kh, kw)).astype(np.float32, order='C')
    weights[0][0][0][0] = 1.89
    weights[1][1][1][1] = 3.0
    weights[2][2][2][2] = 4.0
    x = np.ones((b, c, h, w)).astype(np.float32, order='C')
    ho = (h + 2 * vpadding - vdilation * (kh - 1) - 1) // vstride + 1
    wo = (w + 2 * hpadding - hdilation * (kw - 1) - 1) // hstride + 1
    out = (np.ones((kn, b * ho * wo)) * 10).astype(np.float32, order='C')
    print("Using conv_gemm to compute alpha * weights * im2col(x) + beta * out...")
    conv_gemm = ConvGemm(debug=False)
    conv_gemm_result = conv_gemm.conv_gemm_nchw(weights, x,
                                                vpadding=vpadding, hpadding=hpadding,
                                                vstride=vstride, hstride=hstride,
                                                vdilation=vdilation, hdilation=hdilation,
                                                out=out.reshape(kn, b, ho, wo))
    print(conv_gemm_result)
    print("Sum: ", conv_gemm_result.sum())
    print()
    print("Using im2col and mm...")
    x_c = np.zeros((c * kh * kw, b * ho * wo))
    im2col_nchw_cython(x, x_c,
                       kh, kw, ho, wo,
                       vpadding, hpadding, vstride, hstride, vdilation, hdilation)
    w_c = weights.reshape(kn, -1)
    im2col_mm_result = (w_c @ x_c + out).reshape(kn, b, ho, wo).transpose(1, 0, 2, 3)
    print(im2col_mm_result)
    print("Sum: ", im2col_mm_result.sum())
    print("np.allclose: ", np.allclose(conv_gemm_result, im2col_mm_result))
    # print(conv_gemm_result - im2col_mm_result)
    # Times
    conv_gemm_t = timeit(lambda: conv_gemm.conv_gemm_nchw(weights, x,
                                                          vpadding=vpadding, hpadding=hpadding,
                                                          vstride=vstride, hstride=hstride,
                                                          vdilation=vdilation, hdilation=hdilation),
                         number=10) / 10
    print("Times")
    print("-----")
    print("conv_gemm time: {:.4f}".format(conv_gemm_t))
    im2col_t = timeit(lambda: time_it_func(x, w_c, out, b, kn, ho, wo, kh, kw, vpadding, hpadding, vstride, hstride, vdilation, hdilation),
                      number=10) / 10
    mm_t = timeit(lambda: w_c @ x_c + out, number=10) / 10
    print("im2col+mm time: {:.4f}  (im2col: {:.4f}  mm: {:.4f}".format(im2col_t + mm_t, im2col_t, mm_t))


if __name__ == "__main__":
    __usage_example__()



from pydtnn.layers.conv_2d import Conv2D
from pydtnn.backends.cpu.layers.layer import LayerCPU
from pydtnn.utils.performance_models import im2col_time, matmul_time
from pydtnn.utils.constants import ArrayShape

import numpy as np


class Conv2DCPU(Conv2D[np.ndarray], LayerCPU):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # More parameters initialized in initialize()
        self.biases = None  # type: ignore
        self.weights = None  # type: ignore
        self.fwd_time = None  # type: ignore
        self.bwd_time = None  # type: ignore
    # ----

    def initialize(self, prev_shape: ArrayShape, x: np.ndarray | None = None) -> None:
        super().initialize(prev_shape, x)
        if self.use_bias:
            bias_shape = (self.co,)  # NOTE: Is the same shape in every variant and grouping
            self.biases = self.biases_initializer(bias_shape, self.model.dtype)
            self.db = np.zeros(shape=bias_shape, dtype=self.model.dtype, order="C")

        self.weights = self.weights_initializer(self.weights_shape, self.model.dtype)  # type: ignore (it's ok)
        self.dw: np.ndarray = np.zeros(self.weights.shape, dtype=self.model.dtype, order="C")

        # Performance models
        self.fwd_time = \
            im2col_time(m=(self.ci * self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype) + \
            matmul_time(m=self.co, n=(self.model.batch_size * self.ho * self.wo), k=(self.ci * self.kh * self.kw),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)  # type: ignore (It works well.)
        self.bwd_time = \
            matmul_time(m=self.co, n=(self.ci * self.kh * self.kw), k=(self.model.batch_size * self.ho * self.wo),
                        cpu_speed=self.model.cpu_speed, memory_bw=self.model.memory_bw,
                        dtype=self.model.dtype)
        self.bwd_time += matmul_time(m=(self.ci * self.kh * self.kw), n=(self.model.batch_size * self.ho * self.wo),
                                     k=self.co, cpu_speed=self.model.cpu_speed,
                                     memory_bw=self.model.memory_bw, dtype=self.model.dtype)

    def forward(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Use a real forward variant!")

    def backward(self, dy: np.ndarray) -> np.ndarray:
        raise NotImplementedError("Use a real backwards variant!")

    def print_in_convdirect_format(self) -> None:
        if self.hstride != 1 or self.vstride != 1:
            return
        # #l kn wo ho t kh kw ci wi hi"
        ci, hi, wi = self.model.decode_shape(self.prev_shape)
        print(self.id, self.co, self.wo, self.ho, self.model.batch_size, self.kh, self.kw, ci, wi, hi, sep="\t")



"""
Python interface to the NVIDIA NCCL library
"""

import sys
import ctypes
import ctypes.util
from enum import Enum

if sys.platform in ('linux2', 'linux'):
    _libnccl_libname_list = ['libnccl.so']
elif sys.platform == 'darwin':
    _libnccl_libname_list = ['libnccl.dylib']
elif sys.platform == 'win32':
    _libnccl_libname_list = ['libnccl.dll']
else:
    raise NotImplementedError('PyDTNN NCCL: current platform is not yet supported!')

_libnccl = None
for _libnccl_libname in _libnccl_libname_list:
    try:
        _libnccl = ctypes.cdll.LoadLibrary(_libnccl_libname)
    except OSError:
        pass
    else:
        break
if _libnccl is None:
    raise OSError('NCCL library not found')

# NCCL error
_libnccl.ncclGetErrorString.restype = ctypes.c_char_p
_libnccl.ncclGetErrorString.argtypes = [ctypes.c_int]


class NcclError(Exception):
    def __init__(self, status):
        self.status = status

    def __str__(self):
        error = _libnccl.ncclGetErrorString(self.status)
        return f'{error}'


def ncclCheckStatus(status):
    """
    Raise NCCL exception
    Raise an exception corresponding to the specified NCCL error code.
    Parameters
    ----------
    status : int
        NCCL error code
    """

    if status != 0:
        raise NcclError(status)


NCCL_UNIQUE_ID_BYTES = 128


class NcclUniqueId(ctypes.Structure):
    _fields_ = [("internal", ctypes.c_char * NCCL_UNIQUE_ID_BYTES)]


class NcclComm(ctypes.Structure):
    pass


ncclComm_t = ctypes.POINTER(NcclComm)

# Helper functions

# Return the NCCL_VERSION_CODE of the NCCL library in the supplied integer.
# This integer is coded with the MAJOR, MINOR and PATCH level of the
# NCCL library

# ncclResult_t  ncclGetVersion(int *version);
_libnccl.ncclGetVersion.restype = ctypes.c_size_t
_libnccl.ncclGetVersion.argtypes = []


def ncclGetVersion():
    """
    Get NCCL Version.
    """
    version = ctypes.c_int()
    _libnccl.ncclGetVersion(ctypes.byref(version))
    return version.value


# ncclResult_t  ncclGetUniqueId(NcclUniqueId* uniqueId);
_libnccl.ncclGetUniqueId.restype = int
_libnccl.ncclGetUniqueId.argtypes = []


def ncclGetUniqueId():
    """
    Generates an Id to be used in ncclCommInitRank. ncclGetUniqueId should be
    called once and the Id should be distributed to all ranks in the
    communicator before calling ncclCommInitRank.

    Returns
    -------
    unique_id :
    """

    unique_id = NcclUniqueId()
    status = _libnccl.ncclGetUniqueId(ctypes.byref(unique_id))
    ncclCheckStatus(status)
    return unique_id


# ncclResult_t  ncclCommInitRank(ncclComm_t* comm, int nranks, NcclUniqueId comm_id, int rank);
_libnccl.ncclCommInitRank.restype = int
_libnccl.ncclCommInitRank.argtypes = [ctypes.POINTER(ncclComm_t), ctypes.c_int,
                                      NcclUniqueId, ctypes.c_int]


def ncclCommInitRank(nranks, comm_id, rank):
    """
    Creates a new communicator (multi thread/process version).
    rank must be between 0 and nranks-1 and unique within a communicator clique.
    Each rank is associated to a CUDA device, which has to be set before calling
    ncclCommInitRank.
    ncclCommInitRank implicitly synchronizes with other ranks, so it must be
    called by different threads/processes or use ncclGroupStart/ncclGroupEnd.

    Parameters
    ----------
    nranks :
    comm_id :
    rank :

    Returns
    -------
    comm :
    """

    comm = ncclComm_t()
    status = _libnccl.ncclCommInitRank(ctypes.byref(comm), nranks, comm_id, rank)
    ncclCheckStatus(status)
    return comm


# ncclResult_t  ncclCommInitAll(ncclComm_t* comm, int ndev, const int* dev_list);
_libnccl.ncclCommInitAll.restype = int
_libnccl.ncclCommInitAll.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ctypes.c_int)]


def ncclCommInitAll(dev_list):
    """
    Creates a clique of communicators (single process version).
    This is a convenience function to create a single-process communicator clique.
    Returns an array of ndev newly initialized communicators in comm.
    comm should be pre-allocated with size at least ndev*sizeof(ncclComm_t).
    If dev_list is NULL, the first ndev CUDA devices are used.
    Order of dev_list defines user-order of processors within the communicator.

    Parameters
    ----------
    dev_list :

    Returns
    -------
    comm :
    """
    comm = ncclComm_t() * len(dev_list)
    status = _libnccl.ncclCommInitAll(ctypes.byref(comm), len(dev_list), dev_list)
    ncclCheckStatus(status)
    return comm.value


# ncclResult_t  ncclCommDestroy(ncclComm_t comm);
_libnccl.ncclCommDestroy.restype = int
_libnccl.ncclCommDestroy.argtypes = [ncclComm_t]


def ncclCommDestroy(comm):
    """
    Frees resources associated with communicator object, but waits for any operations
    that might still be running on the device.

    Parameters
    ----------
    comm :
    """
    status = _libnccl.ncclCommDestroy(comm)
    ncclCheckStatus(status)


# ncclResult_t  ncclCommAbort(ncclComm_t comm);
_libnccl.ncclCommAbort.restype = int
_libnccl.ncclCommAbort.argtypes = [ctypes.c_void_p]


def ncclCommAbort(comm):
    """
    Frees resources associated with communicator object, but waits for any operations
    that might still be running on the device.

    Parameters
    ----------
    comm :
    """
    status = _libnccl.ncclCommAbort(comm)
    ncclCheckStatus(status)


# ncclResult_t  ncclCommGetAsyncError(ncclComm_t comm, ncclResult_t *async_error);
_libnccl.ncclCommGetAsyncError.restype = int
_libnccl.ncclCommGetAsyncError.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def ncclCommGetAsyncError(comm):
    """
    Checks whether the comm has encountered any asynchronous errors

    Parameters
    ----------
    comm :
    """
    async_error = ctypes.c_int()
    status = _libnccl.ncclCommGetAsyncError(comm, ctypes.byref(async_error))
    ncclCheckStatus(status)
    return async_error.value


# ncclResult_t  ncclCommCount(const ncclComm_t comm, int* count);
_libnccl.ncclCommCount.restype = int
_libnccl.ncclCommCount.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def ncclCommCount(comm):
    """
    Gets the number of ranks in the communicator clique

    Parameters
    -------
    comm :

    Returns
    -------
    count :
    """

    count = ctypes.c_int()
    status = _libnccl.ncclCommCount(comm, ctypes.byref(count))
    ncclCheckStatus(status)
    return count.value


# ncclResult_t  ncclCommCuDevice(const ncclComm_t comm, int* device);
_libnccl.ncclCommCuDevice.restype = int
_libnccl.ncclCommCuDevice.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def ncclCommCuDevice(comm):
    """
    Returns the cuda device number associated with the communicator.

    Parameters
    -------
    comm :

    Returns
    -------
    device :
    """

    device = ctypes.c_int()
    status = _libnccl.ncclCommCuDevice(comm, ctypes.byref(device))
    ncclCheckStatus(status)
    return device.value


# ncclResult_t  ncclCommUserRank(const ncclComm_t comm, int* rank);
_libnccl.ncclCommUserRank.restype = int
_libnccl.ncclCommUserRank.argtypes = [ctypes.c_void_p, ctypes.c_void_p]


def ncclCommUserRank(comm):
    """
    Returns the user-ordered "rank" associated with the communicator.

    Parameters
    -------
    comm :

    Returns
    -------
    rank :
    """

    rank = ctypes.c_int()
    status = _libnccl.ncclCommUserRank(comm, ctypes.byref(rank))
    ncclCheckStatus(status)
    return rank.value


# Reduction operation selector
class RedOp(Enum):
    Sum = 0
    Prod = 1
    Max = 2
    Min = 3
    NumOps = 4


# Data types
class DataType(Enum):
    Int8 = Char = 0
    Uint8 = 1
    Int32 = Int = 2
    Uint32 = 3
    Int64 = 4
    Uint64 = 5
    Float16 = Half = 6
    Float32 = Float = 7
    Float64 = Double = 8
    NumTypes = 9


# Collective communication operations
#
# Collective communication operations must be called separately for each
# communicator in a communicator clique.
#
# They return when operations have been enqueued on the CUDA stream.
#
# Since they may perform inter-CPU synchronization, each call has to be done
# from a different thread or process, or need to use Group Semantics (see
# below).


# ncclResult_t  ncclReduce(const void* sendbuff, void* recvbuff, size_t count, ncclDataType_t datatype,
#     ncclRedOp_t op, int root, ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclReduce.restype = int
_libnccl.ncclReduce.argtypes = [ctypes.c_void_p, ctypes.c_void_p,
                                ctypes.c_ulong, ctypes.c_int, ctypes.c_int, ctypes.c_int,
                                ncclComm_t, ctypes.c_void_p]


def ncclReduce(sendbuff, recvbuff, count, datatype, op, root, comm, stream):
    """
    Reduces data arrays of length count in sendbuff into recvbuff using op operation.
    recvbuff may be NULL on all calls except for root device.
    root is the rank (not the CUDA device) where data will reside after the
    operation is complete.

     In-place operation will happen if sendbuff == recvbuff.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclReduce(sendbuff, recvbuff, count,
                                 datatype.value, op.value, root, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclBroadcast(const void* sendbuff, void* recvbuff, size_t count, ncclDataType_t datatype, int root,
#     ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclBroadcast.restype = int
_libnccl.ncclBroadcast.argtypes = [ctypes.c_void_p, ctypes.c_void_p,
                                   ctypes.c_ulong, ctypes.c_int, ctypes.c_int,
                                   ncclComm_t, ctypes.c_void_p]


def ncclBroadcast(sendbuff, recvbuff, count, datatype, root, comm, stream):
    """
    Copies count values from root to all other devices.
    root is the rank (not the CUDA device) where data resides before the
    operation is started.

     In-place operation will happen if sendbuff == recvbuff.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclBroadcast(sendbuff, recvbuff, count, datatype.value, root, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclAllReduce(const void* sendbuff, void* recvbuff, size_t count,
#     ncclDataType_t datatype, ncclRedOp_t op, ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclAllReduce.restype = int
_libnccl.ncclAllReduce.argtypes = [ctypes.c_void_p, ctypes.c_void_p,
                                   ctypes.c_ulong, ctypes.c_int, ctypes.c_int,
                                   ncclComm_t, ctypes.c_void_p]


def ncclAllReduce(sendbuff, recvbuff, count, datatype, op, comm, stream):
    """
    Reduces data arrays of length count in sendbuff using op operation, and
    leaves identical copies of result on each recvbuff.

    In-place operation will happen if sendbuff == recvbuff.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclAllReduce(sendbuff, recvbuff, count,
                                    datatype.value, op.value, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclReduceScatter(const void* sendbuff, void* recvbuff,
#     size_t recvcount, ncclDataType_t datatype, ncclRedOp_t op, ncclComm_t comm,
#     cudaStream_t stream);
_libnccl.ncclReduceScatter.restype = int
_libnccl.ncclReduceScatter.argtypes = [ctypes.c_void_p, ctypes.c_void_p,
                                       ctypes.c_ulong, ctypes.c_int,
                                       ncclComm_t, ctypes.c_void_p]


def ncclReduceScatter(sendbuff, recvbuff, recvcount, datatype, comm, stream):
    """
    Reduces data in sendbuff using op operation and leaves reduced result
    scattered over the devices so that recvbuff on rank i will contain the i-th
    block of the result.
    Assumes sendcount is equal to nranks*recvcount, which means that sendbuff
    should have a size of at least nranks*recvcount elements.

    In-place operations will happen if recvbuff == sendbuff + rank * recvcount.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclReduceScatter(sendbuff, recvbuff, recvcount,
                                        datatype.value, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclAllGather(const void* sendbuff, void* recvbuff, size_t sendcount,
#     ncclDataType_t datatype, ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclAllGather.restype = int
_libnccl.ncclAllGather.argtypes = [ctypes.c_void_p, ctypes.c_void_p,
                                   ctypes.c_ulong, ctypes.c_int,
                                   ncclComm_t, ctypes.c_void_p]


def ncclAllGather(sendbuff, recvbuff, sendcount, datatype, comm, stream):
    """
    Each device gathers sendcount values from other GPUs into recvbuff,
    receiving data from rank i at offset i*sendcount.
    Assumes recvcount is equal to nranks*sendcount, which means that recvbuff
    should have a size of at least nranks*sendcount elements.

    In-place operations will happen if sendbuff == recvbuff + rank * sendcount.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclAllGather(sendbuff, recvbuff, sendcount,
                                    datatype.value, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclSend(const void* sendbuff, size_t count, ncclDataType_t datatype, int peer,
#     ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclSend.restype = int
_libnccl.ncclSend.argtypes = [ctypes.c_void_p,
                              ctypes.c_ulong, ctypes.c_int, ctypes.c_int,
                              ncclComm_t, ctypes.c_void_p]


def ncclSend(sendbuff, count, datatype, peer, comm, stream):
    """
    Send data from sendbuff to rank peer.

    Rank peer needs to call ncclRecv with the same datatype and the same count from this
    rank.

    This operation is blocking for the GPU. If multiple ncclSend and ncclRecv operations
    need to progress concurrently to complete, they must be fused within a ncclGroupStart/
    ncclGroupEnd section.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclSend(sendbuff, count, datatype.value, peer, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclRecv(void* recvbuff, size_t count, ncclDataType_t datatype, int peer,
#    ncclComm_t comm, cudaStream_t stream);
_libnccl.ncclRecv.restype = int
_libnccl.ncclRecv.argtypes = [ctypes.c_void_p,
                              ctypes.c_int, ctypes.c_int, ctypes.c_int,
                              ncclComm_t, ctypes.c_void_p]


def ncclRecv(recvbuff, count, datatype, peer, comm, stream):
    """
    Receive data from rank peer into recvbuff.

    Rank peer needs to call ncclSend with the same datatype and the same count to this
    rank.

    This operation is blocking for the GPU. If multiple ncclSend and ncclRecv operations
    need to progress concurrently to complete, they must be fused within a ncclGroupStart/
    ncclGroupEnd section.

    Parameters
    -----------

    Returns
    -------

    """

    status = _libnccl.ncclRecv(recvbuff, count, datatype.value, peer, comm, stream)
    ncclCheckStatus(status)


# ncclResult_t  ncclGroupStart();
_libnccl.ncclGroupStart.restype = int
_libnccl.ncclGroupStart.argtypes = []


def ncclGroupStart():
    """
    Start a group call. All calls to NCCL until ncclGroupEnd will be fused into
    a single NCCL operation. Nothing will be started on the CUDA stream until
    ncclGroupEnd.

    Returns
    -------

    """

    status = _libnccl.ncclGroupStart()
    ncclCheckStatus(status)


# ncclResult_t  ncclGroupEnd();
_libnccl.ncclGroupEnd.restype = int
_libnccl.ncclGroupEnd.argtypes = []


def ncclGroupEnd():
    """
    End a group call. Start a fused NCCL operation consisting of all calls since
    ncclGroupStart. Operations on the CUDA stream depending on the NCCL operations
    need to be called after ncclGroupEnd.

    Returns
    -------

    """

    status = _libnccl.ncclGroupEnd()
    ncclCheckStatus(status)



from typing import TYPE_CHECKING

import numpy as np

from pydtnn.optimizers.optimizer import Optimizer
from pydtnn.utils.constants import Array

if TYPE_CHECKING:
    from pydtnn.model import Model


class SGD[T: Array](Optimizer[T]):
    """
    SGD Optimizer
    """

    def __init__(self, learning_rate: float = 1e-2, momentum: float = 0.9,
                 nesterov: bool = False, decay: float = 0.0, dtype: np.dtype = np.dtype(np.float32)):
        super().__init__(learning_rate=learning_rate, dtype=dtype)
        self.momentum: float = momentum
        self.nesterov: bool = nesterov
        self.decay: float = decay

    @classmethod
    def from_model(cls, model: "Model") -> "SGD":
        return SGD(learning_rate=model.learning_rate,
                   momentum=model.optimizer_momentum,
                   nesterov=model.optimizer_nesterov,
                   decay=model.optimizer_decay,
                   dtype=model.dtype)



from pydtnn.backends.cpu.metrics.metric import MetricCPU
from pydtnn.metrics.binary_confusion_matrix import BinaryConfusionMatrix

import numpy as np

TRUE_POSITIVE = (0, 0)
TRUE_NEGATIVE = (1, 1)
FALSE_NEGATIVE = (0, 1)
FALSE_POSITIVE = (1, 0)

_dict_indexes = {
    # y_targ[i, label] == y_pred[i, label]
    # i.e.: "are both target' and prediciton' values the same?"
    True:{
        # bool(y_targ[i, label])
        # i.e.: "is the target' value 1 (True) or 0 (False)?"
        True: TRUE_POSITIVE,
        False: TRUE_NEGATIVE
    },
    False:{
        # bool(y_targ[i, label])
        True: FALSE_NEGATIVE,
        False: FALSE_POSITIVE
    }
}

class BinaryConfusionMatrixCPU(BinaryConfusionMatrix[np.ndarray], MetricCPU):

    def initialize(self) -> None:
        super().initialize()

    def compute(self, y_pred: np.ndarray, y_targ: np.ndarray) -> np.ndarray:
        """
        For every label in target class, the output will have one confusion matrix like this:
                |Predicted|
        ________| T  | F  |
        Target|T| TP | FN |
              |F| FP | TN |
        """
        
        # NOTE: y_pred.shape == y_targ.shape == (n<=self.model.batch_size, self.model.output_shape)
        n, target_classes = y_pred.shape
        #assert target_classes == pred_classes, f"target_classes ({target_classes}) != pred_classes {pred_classes}, and must have the same value."
        conf_matrix = np.zeros((target_classes, 2, 2), dtype=np.int32)

        for i in range(n):
            for label in range(target_classes):
                conf_matrix[label, *(_dict_indexes[y_targ[i, label] == y_pred[i, label]] [bool(y_targ[i, label])])] += 1

        self.conf_matrix = conf_matrix

        return self.conf_matrix

    def get_true_positives(self): 
        return self.conf_matrix[:, *TRUE_POSITIVE]

    def get_true_negatives(self): 
        return self.conf_matrix[:, *TRUE_NEGATIVE]

    def get_false_positives(self): 
        return self.conf_matrix[:, *FALSE_NEGATIVE]

    def get_false_negatives(self): 
        return self.conf_matrix[:, *FALSE_POSITIVE]


